import math
from datetime import date, timedelta

from ib_insync import Option, Order, PortfolioItem
from ib_insync.contract import Stock

from thetagang.util import (
    calculate_net_short_positions,
    get_target_delta,
    position_pnl,
    weighted_avg_long_strike,
    weighted_avg_short_strike,
    would_increase_spread,
)


def test_position_pnl() -> None:
    qqq_put = PortfolioItem(
        contract=Option(
            conId=397556522,
            symbol="QQQ",
            lastTradeDateOrContractMonth="20201218",
            strike=300.0,
            right="P",
            multiplier="100",
            primaryExchange="AMEX",
            currency="USD",
            localSymbol="QQQ   201218P00300000",
            tradingClass="QQQ",
        ),
        position=-1.0,
        marketPrice=4.1194396,
        marketValue=-411.94,
        averageCost=222.4293,
        unrealizedPNL=-189.51,
        realizedPNL=0.0,
        account="DU2962946",
    )
    assert round(position_pnl(qqq_put), 2) == -0.85

    spy = PortfolioItem(
        contract=Stock(
            conId=756733,
            symbol="SPY",
            right="0",
            primaryExchange="ARCA",
            currency="USD",
            localSymbol="SPY",
            tradingClass="SPY",
        ),
        position=100.0,
        marketPrice=365.4960022,
        marketValue=36549.6,
        averageCost=368.42,
        unrealizedPNL=-292.4,
        realizedPNL=0.0,
        account="DU2962946",
    )
    assert round(position_pnl(spy), 4) == -0.0079

    spy_call = PortfolioItem(
        contract=Option(
            conId=454208258,
            symbol="SPY",
            lastTradeDateOrContractMonth="20201214",
            strike=373.0,
            right="C",
            multiplier="100",
            primaryExchange="AMEX",
            currency="USD",
            localSymbol="SPY   201214C00373000",
            tradingClass="SPY",
        ),
        position=-1.0,
        marketPrice=0.08,
        marketValue=-8.0,
        averageCost=96.422,
        unrealizedPNL=88.42,
        realizedPNL=0.0,
        account="DU2962946",
    )
    assert round(position_pnl(spy_call), 2) == 0.92

    spy_put = PortfolioItem(
        contract=Option(
            conId=458705534,
            symbol="SPY",
            lastTradeDateOrContractMonth="20210122",
            strike=352.5,
            right="P",
            multiplier="100",
            primaryExchange="AMEX",
            currency="USD",
            localSymbol="SPY   210122P00352500",
            tradingClass="SPY",
        ),
        position=-1.0,
        marketPrice=5.96710015,
        marketValue=-596.71,
        averageCost=528.9025,
        unrealizedPNL=-67.81,
        realizedPNL=0.0,
        account="DU2962946",
    )
    assert round(position_pnl(spy_put), 2) == -0.13


def test_get_delta() -> None:
    config = {"target": {"delta": 0.5}, "symbols": {"SPY": {"weight": 1}}}
    assert 0.5 == get_target_delta(config, "SPY", "P")

    config = {
        "target": {"delta": 0.5, "puts": {"delta": 0.4}},
        "symbols": {"SPY": {"weight": 1}},
    }
    assert 0.4 == get_target_delta(config, "SPY", "P")

    config = {
        "target": {"delta": 0.5, "calls": {"delta": 0.4}},
        "symbols": {"SPY": {"weight": 1}},
    }
    assert 0.5 == get_target_delta(config, "SPY", "P")

    config = {
        "target": {"delta": 0.5, "calls": {"delta": 0.4}},
        "symbols": {"SPY": {"weight": 1}},
    }
    assert 0.4 == get_target_delta(config, "SPY", "C")

    config = {
        "target": {"delta": 0.5, "calls": {"delta": 0.4}},
        "symbols": {"SPY": {"weight": 1, "delta": 0.3}},
    }
    assert 0.3 == get_target_delta(config, "SPY", "C")

    config = {
        "target": {"delta": 0.5, "calls": {"delta": 0.4}},
        "symbols": {"SPY": {"weight": 1, "delta": 0.3, "puts": {"delta": 0.2}}},
    }
    assert 0.3 == get_target_delta(config, "SPY", "C")

    config = {
        "target": {"delta": 0.5, "calls": {"delta": 0.4}},
        "symbols": {"SPY": {"weight": 1, "delta": 0.3, "puts": {"delta": 0.2}}},
    }
    assert 0.2 == get_target_delta(config, "SPY", "P")


def con(dte: str, strike: float, right: str, position: float) -> PortfolioItem:
    return PortfolioItem(
        contract=Option(
            conId=458705534,
            symbol="SPY",
            lastTradeDateOrContractMonth=dte,
            strike=strike,
            right=right,
            multiplier="100",
            primaryExchange="AMEX",
            currency="USD",
            localSymbol="SPY   210122P00352500",
            tradingClass="SPY",
        ),
        position=position,
        marketPrice=5.96710015,
        marketValue=-596.71,
        averageCost=528.9025,
        unrealizedPNL=-67.81,
        realizedPNL=0.0,
        account="DU2962946",
    )


def test_calculate_net_short_positions() -> None:
    today = date.today()
    exp3dte = (today + timedelta(days=3)).strftime("%Y%m%d")
    exp30dte = (today + timedelta(days=30)).strftime("%Y%m%d")
    exp90dte = (today + timedelta(days=90)).strftime("%Y%m%d")

    assert 1 == calculate_net_short_positions([con(exp3dte, 69, "P", -1)], "P")

    assert 1 == calculate_net_short_positions(
        [con(exp3dte, 69, "P", -1), con(exp3dte, 69, "C", 1)], "P"
    )

    assert 0 == calculate_net_short_positions(
        [con(exp3dte, 69, "P", -1), con(exp3dte, 69, "C", 1)], "C"
    )

    assert 0 == calculate_net_short_positions(
        [con(exp3dte, 69, "C", -1), con(exp3dte, 69, "C", 1)], "C"
    )

    assert 0 == calculate_net_short_positions(
        [
            con(exp3dte, 69, "C", -1),
            con(exp3dte, 69, "C", 1),
            con(exp30dte, 69, "C", 1),
        ],
        "C",
    )

    assert 0 == calculate_net_short_positions(
        [
            con(exp3dte, 69, "C", -1),
            con(exp3dte, 69, "P", -1),
            con(exp3dte, 69, "C", 1),
            con(exp30dte, 69, "C", 1),
        ],
        "C",
    )

    assert 0 == calculate_net_short_positions(
        [
            con(exp3dte, 69, "C", -1),
            con(exp3dte, 69, "P", -1),
            con(exp3dte, 69, "C", 1),
            con(exp30dte, 70, "C", 1),
        ],
        "C",
    )

    assert 1 == calculate_net_short_positions(
        [
            con(exp3dte, 69, "C", -1),
            con(exp3dte, 69, "C", -1),
            con(exp3dte, 69, "C", 1),
            con(exp30dte, 70, "C", 1),
        ],
        "C",
    )

    assert 2 == calculate_net_short_positions(
        [
            con(exp3dte, 69, "C", -1),
            con(exp3dte, 69, "C", -1),
            con(exp3dte, 69, "P", 1),
            con(exp30dte, 69, "P", 1),
        ],
        "C",
    )

    assert 0 == calculate_net_short_positions(
        [
            con(exp3dte, 69, "C", -1),
            con(exp3dte, 69, "C", -1),
            con(exp3dte, 69, "C", 1),
            con(exp30dte, 69, "C", 5),
        ],
        "C",
    )

    assert 0 == calculate_net_short_positions(
        [
            con(exp3dte, 69, "C", -1),
            con(exp30dte, 69, "C", -1),
            con(exp3dte, 69, "C", 1),
            con(exp30dte, 69, "C", 5),
        ],
        "C",
    )

    assert 0 == calculate_net_short_positions(
        [
            con(exp3dte, 69, "P", -1),
            con(exp30dte, 69, "P", -1),
            con(exp3dte, 69, "P", 1),
            con(exp30dte, 69, "P", 5),
        ],
        "P",
    )

    assert 0 == calculate_net_short_positions(
        [
            con(exp3dte, 70, "P", -1),
            con(exp30dte, 69, "P", -1),
            con(exp3dte, 69, "P", 1),
            con(exp30dte, 70, "P", 5),
        ],
        "P",
    )

    assert 2 == calculate_net_short_positions(
        [
            con(exp3dte, 70, "P", -1),
            con(exp30dte, 69, "P", -1),
            con(exp3dte, 69, "P", 1),
            con(exp30dte, 68, "P", 5),
        ],
        "P",
    )

    assert 0 == calculate_net_short_positions(
        [
            con(exp3dte, 70, "C", -1),
            con(exp30dte, 69, "C", -1),
            con(exp3dte, 69, "C", 1),
            con(exp30dte, 68, "C", 5),
        ],
        "C",
    )

    assert 1 == calculate_net_short_positions(
        [
            con(exp3dte, 70, "C", -1),
            con(exp30dte, 69, "C", -1),
            con(exp3dte, 71, "C", 1),
            con(exp30dte, 70, "C", 5),
        ],
        "C",
    )

    assert 2 == calculate_net_short_positions(
        [
            con(exp3dte, 70, "C", -1),
            con(exp30dte, 71, "C", -1),
            con(exp3dte, 71, "C", 1),
            con(exp30dte, 72, "C", 5),
        ],
        "C",
    )

    assert 3 == calculate_net_short_positions(
        [
            con(exp3dte, 70, "C", -1),
            con(exp30dte, 71, "C", -1),
            con(exp90dte, 72, "C", -1),
            con(exp3dte, 71, "C", 1),
            con(exp30dte, 72, "C", 5),
        ],
        "C",
    )

    assert 5 == calculate_net_short_positions(
        [
            con(exp3dte, 60, "P", -10),
            con(exp30dte, 69, "P", -1),
            con(exp90dte, 69, "P", 1),
            con(exp90dte, 68, "P", 5),
        ],
        "P",
    )

    assert 10 == calculate_net_short_positions(
        [
            con(exp3dte, 70, "P", -10),
            con(exp30dte, 69, "P", -1),
            con(exp90dte, 69, "P", 1),
            con(exp90dte, 68, "P", 5),
        ],
        "P",
    )

    assert 0 == calculate_net_short_positions(
        [
            con(exp3dte, 60, "P", -10),
            con(exp30dte, 69, "P", -1),
            con(exp90dte, 69, "P", 1),
            con(exp90dte, 68, "P", 50),
        ],
        "P",
    )

    # A couple real-world examples
    exp9dte = (today + timedelta(days=9)).strftime("%Y%m%d")
    exp16dte = (today + timedelta(days=16)).strftime("%Y%m%d")
    exp23dte = (today + timedelta(days=23)).strftime("%Y%m%d")
    exp30dte = (today + timedelta(days=30)).strftime("%Y%m%d")
    exp37dte = (today + timedelta(days=37)).strftime("%Y%m%d")
    exp268dte = (today + timedelta(days=268)).strftime("%Y%m%d")

    assert 2 == calculate_net_short_positions(
        [
            con(exp9dte, 77.0, "P", -2),
            con(exp16dte, 76.0, "P", -1),
            con(exp16dte, 77.0, "P", -1),
            con(exp23dte, 77.0, "P", -6),
            con(exp30dte, 77.0, "P", -2),
            con(exp37dte, 77.0, "P", -5),
            con(exp268dte, 77.0, "P", 15),
        ],
        "P",
    )

    assert 0 == calculate_net_short_positions(
        [
            con(exp9dte, 77.0, "P", -2),
            con(exp16dte, 76.0, "P", -1),
            con(exp16dte, 77.0, "P", -1),
            con(exp23dte, 77.0, "P", -6),
            con(exp30dte, 77.0, "P", -2),
            con(exp37dte, 77.0, "P", -5),
            con(exp268dte, 77.0, "P", 15),
        ],
        "C",
    )

    assert 20 == calculate_net_short_positions(
        [
            con(exp23dte, 72.0, "C", -8),
            con(exp30dte, 66.0, "C", -8),
            con(exp30dte, 68.0, "C", -9),
            con(exp30dte, 69.0, "C", -7),
            con(exp30dte, 72.0, "C", -1),
            con(exp37dte, 59.5, "C", -8),
            con(exp37dte, 68.0, "C", -7),
            con(exp268dte, 55.0, "C", 5),
            con(exp268dte, 60.0, "C", 23),
        ],
        "C",
    )

    assert 0 == calculate_net_short_positions(
        [
            con(exp23dte, 72.0, "C", -8),
            con(exp30dte, 66.0, "C", -8),
            con(exp30dte, 68.0, "C", -9),
            con(exp30dte, 69.0, "C", -7),
            con(exp30dte, 72.0, "C", -1),
            con(exp37dte, 59.5, "C", -8),
            con(exp37dte, 68.0, "C", -7),
            con(exp268dte, 55.0, "C", 5),
            con(exp268dte, 60.0, "C", 23),
        ],
        "P",
    )


def test_weighted_avg_strike() -> None:
    today = date.today()
    exp3dte = (today + timedelta(days=3)).strftime("%Y%m%d")
    exp30dte = (today + timedelta(days=30)).strftime("%Y%m%d")
    exp90dte = (today + timedelta(days=90)).strftime("%Y%m%d")

    assert math.isclose(
        70,
        weighted_avg_short_strike(
            [
                con(exp3dte, 70, "C", -1),
                con(exp30dte, 70, "C", -1),
                con(exp90dte, 70, "C", -1),
                con(exp3dte, 100, "C", 1),
                con(exp30dte, 100, "C", 5),
            ],
            "C",
        )
        or -1,
    )
    assert math.isclose(
        100,
        weighted_avg_long_strike(
            [
                con(exp3dte, 70, "C", -1),
                con(exp30dte, 70, "C", -1),
                con(exp90dte, 70, "C", -1),
                con(exp3dte, 100, "C", 1),
                con(exp30dte, 100, "C", 5),
            ],
            "C",
        )
        or -1,
    )
    assert math.isclose(
        70,
        weighted_avg_short_strike(
            [
                con(exp3dte, 70, "P", -1),
                con(exp30dte, 70, "P", -1),
                con(exp90dte, 70, "P", -1),
                con(exp3dte, 100, "P", 1),
                con(exp30dte, 100, "P", 5),
            ],
            "P",
        )
        or -1,
    )
    assert math.isclose(
        100,
        weighted_avg_long_strike(
            [
                con(exp3dte, 70, "P", -1),
                con(exp30dte, 70, "P", -1),
                con(exp90dte, 70, "P", -1),
                con(exp3dte, 100, "P", 1),
                con(exp30dte, 100, "P", 5),
            ],
            "P",
        )
        or -1,
    )

    assert math.isclose(
        28,
        weighted_avg_short_strike(
            [
                con(exp3dte, 10, "P", -4),
                con(exp3dte, 100, "P", -1),
                con(exp3dte, 100, "P", 4),
                con(exp3dte, 10, "P", 1),
            ],
            "P",
        )
        or -1,
    )

    assert math.isclose(
        82,
        weighted_avg_long_strike(
            [
                con(exp3dte, 10, "P", -4),
                con(exp3dte, 100, "P", -1),
                con(exp3dte, 100, "P", 4),
                con(exp3dte, 10, "P", 1),
            ],
            "P",
        )
        or -1,
    )


def test_would_increase_spread() -> None:
    # Test BUY order with lmtPrice < 0 and updated_price > lmtPrice
    order1 = Order(action="BUY", lmtPrice=-10)
    updated_price1 = -5.0
    assert would_increase_spread(order1, updated_price1) is False

    # Test BUY order with lmtPrice < 0 and updated_price < lmtPrice
    order2 = Order(action="BUY", lmtPrice=-10)
    updated_price2 = -15.0
    assert would_increase_spread(order2, updated_price2) is True

    # Test BUY order with lmtPrice > 0 and updated_price < lmtPrice
    order3 = Order(action="BUY", lmtPrice=10)
    updated_price3 = 5.0
    assert would_increase_spread(order3, updated_price3) is True

    # Test BUY order with lmtPrice > 0 and updated_price > lmtPrice
    order4 = Order(action="BUY", lmtPrice=10)
    updated_price4 = 15.0
    assert would_increase_spread(order4, updated_price4) is False

    # Test SELL order with lmtPrice < 0 and updated_price < lmtPrice
    order5 = Order(action="SELL", lmtPrice=-10)
    updated_price5 = -15.0
    assert would_increase_spread(order5, updated_price5) is False

    # Test SELL order with lmtPrice < 0 and updated_price > lmtPrice
    order6 = Order(action="SELL", lmtPrice=-10)
    updated_price6 = -5.0
    assert would_increase_spread(order6, updated_price6) is True

    # Test SELL order with lmtPrice > 0 and updated_price > lmtPrice
    order7 = Order(action="SELL", lmtPrice=10)
    updated_price7 = 15.0
    assert would_increase_spread(order7, updated_price7) is True

    # Test SELL order with lmtPrice > 0 and updated_price < lmtPrice
    order8 = Order(action="SELL", lmtPrice=10)
    updated_price8 = 5.0
    assert would_increase_spread(order8, updated_price8) is False



from asyncio import Future

from ib_insync import IB, IBC, Watchdog, util
from ib_insync.contract import Contract
from rich import box
from rich.console import Console, Group
from rich.panel import Panel
from rich.table import Table
from rich.tree import Tree

from thetagang.config import normalize_config, validate_config
from thetagang.fmt import dfmt, ffmt, pfmt
from thetagang.util import (
    get_strike_limit,
    get_target_delta,
    get_write_threshold_perc,
    get_write_threshold_sigma,
    maintain_high_water_mark,
)

from .portfolio_manager import PortfolioManager

util.patchAsyncio()

console = Console()


def start(config_path: str, without_ibc: bool = False) -> None:
    import toml

    with open(config_path, "r", encoding="utf8") as file:
        config = toml.load(file)

    config = normalize_config(config)

    validate_config(config)

    config_table = Table(box=box.SIMPLE_HEAVY)
    config_table.add_column("Section")
    config_table.add_column("Setting")
    config_table.add_column("")
    config_table.add_column("Value")

    config_table.add_row("[spring_green1]Account details")
    config_table.add_row("", "Account number", "=", config["account"]["number"])
    config_table.add_row(
        "", "Cancel existing orders", "=", f'{config["account"]["cancel_orders"]}'
    )
    config_table.add_row(
        "",
        "Margin usage",
        "=",
        f"{config['account']['margin_usage']} ({pfmt(config['account']['margin_usage'],0)})",
    )
    config_table.add_row(
        "", "Market data type", "=", f'{config["account"]["market_data_type"]}'
    )

    config_table.add_section()
    config_table.add_row("[spring_green1]Constants")
    config_table.add_row(
        "",
        "Daily stddev window",
        "=",
        f"{config['constants']['daily_stddev_window']}",
    )

    c_write_thresh = (
        f"{ffmt(get_write_threshold_sigma(config, None, 'C'))}σ"
        if get_write_threshold_sigma(config, None, "C")
        else pfmt(get_write_threshold_perc(config, None, "C"))
    )
    p_write_thresh = (
        f"{ffmt(get_write_threshold_sigma(config, None, 'P'))}σ"
        if get_write_threshold_sigma(config, None, "P")
        else pfmt(get_write_threshold_perc(config, None, "P"))
    )

    config_table.add_row("", "Write threshold for puts", "=", p_write_thresh)
    config_table.add_row("", "Write threshold for calls", "=", c_write_thresh)

    config_table.add_section()
    config_table.add_row("[spring_green1]Order settings")
    config_table.add_row(
        "",
        "Exchange",
        "=",
        f"{config['orders']['exchange']}",
    )
    config_table.add_row(
        "",
        "Params",
        "=",
        f"{config['orders']['algo']['params']}",
    )
    config_table.add_row(
        "",
        "Price update delay",
        "=",
        f"{config['orders']['price_update_delay']}",
    )
    config_table.add_row(
        "",
        "Minimum credit",
        "=",
        f"{dfmt(config['orders']['minimum_credit'])}",
    )

    config_table.add_section()
    config_table.add_row("[spring_green1]Close option positions")
    config_table.add_row(
        "",
        "When P&L",
        ">=",
        f"{pfmt(config['roll_when']['close_at_pnl'],0)}",
    )

    config_table.add_section()
    config_table.add_row("[spring_green1]Roll options when either condition is true")
    config_table.add_row(
        "",
        "Days to expiry",
        "<=",
        f"{config['roll_when']['dte']} and P&L >= {config['roll_when']['min_pnl']} ({pfmt(config['roll_when']['min_pnl'],0)})",
    )
    if "max_dte" in config["roll_when"]:
        config_table.add_row(
            "",
            "P&L",
            ">=",
            f"{config['roll_when']['pnl']} ({pfmt(config['roll_when']['pnl'],0)}) and DTE < {config['roll_when']['max_dte']}",
        )
    else:
        config_table.add_row(
            "",
            "P&L",
            ">=",
            f"{config['roll_when']['pnl']} ({pfmt(config['roll_when']['pnl'],0)})",
        )

    config_table.add_row(
        "",
        "Puts: credit only",
        "=",
        f"{config['roll_when']['puts']['credit_only']}",
    )
    config_table.add_row(
        "",
        "Puts: roll excess",
        "=",
        f"{config['roll_when']['puts']['has_excess']}",
    )
    config_table.add_row(
        "",
        "Calls: credit only",
        "=",
        f"{config['roll_when']['calls']['credit_only']}",
    )
    config_table.add_row(
        "",
        "Calls: roll excess",
        "=",
        f"{config['roll_when']['calls']['has_excess']}",
    )
    config_table.add_row(
        "",
        "Calls: maintain high water mark",
        "=",
        f"{config['roll_when']['calls']['maintain_high_water_mark']}",
    )
    config_table.add_section()
    config_table.add_row("[spring_green1]When writing new contracts")
    config_table.add_row(
        "",
        "Calculate net contract positions",
        "=",
        f"{config['write_when']['calculate_net_contracts']}",
    )
    config_table.add_row(
        "",
        "Puts, write when red",
        "=",
        f"{config['write_when']['puts']['red']}",
    )
    config_table.add_row(
        "",
        "Puts, write when green",
        "=",
        f"{config['write_when']['puts']['green']}",
    )
    config_table.add_row(
        "",
        "Calls, write when green",
        "=",
        f"{config['write_when']['calls']['green']}",
    )
    config_table.add_row(
        "",
        "Calls, write when red",
        "=",
        f"{config['write_when']['calls']['red']}",
    )
    config_table.add_row(
        "",
        "Call cap factor",
        "=",
        f"{pfmt(config['write_when']['calls']['cap_factor'])}",
    )
    config_table.add_row(
        "",
        "Call cap target floor",
        "=",
        f"{pfmt(config['write_when']['calls']['cap_target_floor'])}",
    )

    config_table.add_section()
    config_table.add_row("[spring_green1]When contracts are ITM")
    config_table.add_row(
        "",
        "Roll puts",
        "=",
        f"{config['roll_when']['puts']['itm']}",
    )
    config_table.add_row(
        "",
        "Roll puts always",
        "=",
        f"{config['roll_when']['puts']['always_when_itm']}",
    )
    config_table.add_row(
        "",
        "Roll calls",
        "=",
        f"{config['roll_when']['calls']['itm']}",
    )
    config_table.add_row(
        "",
        "Roll calls always",
        "=",
        f"{config['roll_when']['calls']['always_when_itm']}",
    )

    config_table.add_section()
    config_table.add_row("[spring_green1]Write options with targets of")
    config_table.add_row("", "Days to expiry", ">=", f"{config['target']['dte']}")
    if "max_dte" in config["target"]:
        config_table.add_row(
            "", "Days to expiry", "<=", f"{config['target']['max_dte']}"
        )
    config_table.add_row("", "Default delta", "<=", f"{config['target']['delta']}")
    if "puts" in config["target"]:
        config_table.add_row(
            "",
            "Delta for puts",
            "<=",
            f"{config['target']['puts']['delta']}",
        )
    if "calls" in config["target"]:
        config_table.add_row(
            "",
            "Delta for calls",
            "<=",
            f"{config['target']['calls']['delta']}",
        )
    config_table.add_row(
        "",
        "Maximum new contracts",
        "=",
        f"{pfmt(config['target']['maximum_new_contracts_percent'],0)} of buying power",
    )
    config_table.add_row(
        "",
        "Minimum open interest",
        "=",
        f"{config['target']['minimum_open_interest']}",
    )

    config_table.add_section()
    config_table.add_row("[spring_green1]Cash management")
    config_table.add_row("", "Enabled", "=", f"{config['cash_management']['enabled']}")
    config_table.add_row(
        "", "Cash fund", "=", f"{config['cash_management']['cash_fund']}"
    )
    config_table.add_row(
        "",
        "Target cash",
        "=",
        f"{dfmt(config['cash_management']['target_cash_balance'])}",
    )
    config_table.add_row(
        "",
        "Buy threshold",
        "=",
        f"{dfmt(config['cash_management']['buy_threshold'])}",
    )
    config_table.add_row(
        "",
        "Sell threshold",
        "=",
        f"{dfmt(config['cash_management']['sell_threshold'])}",
    )

    config_table.add_section()
    config_table.add_row("[spring_green1]Hedging with VIX calls")
    config_table.add_row("", "Enabled", "=", f"{config['vix_call_hedge']['enabled']}")
    config_table.add_row(
        "",
        "Target delta",
        "<=",
        f"{config['vix_call_hedge']['delta']}",
    )
    config_table.add_row(
        "",
        "Target DTE",
        ">=",
        f"{config['vix_call_hedge']['target_dte']}",
    )
    config_table.add_row(
        "",
        "Ignore DTE",
        "<=",
        f"{config['vix_call_hedge']['ignore_dte']}",
    )
    config_table.add_row(
        "",
        "Ignore DTE",
        "<=",
        f"{config['vix_call_hedge']['ignore_dte']}",
    )
    config_table.add_row(
        "",
        "Close hedges when VIX",
        ">=",
        f"{config['vix_call_hedge']['close_hedges_when_vix_exceeds']}",
    )
    for alloc in config["vix_call_hedge"]["allocation"]:
        config_table.add_row()
        if "lower_bound" in alloc:
            config_table.add_row(
                "",
                f"Allocate {pfmt(alloc['weight'])} when VIXMO",
                ">=",
                f"{alloc['lower_bound']}",
            )
        if "upper_bound" in alloc:
            config_table.add_row(
                "",
                f"Allocate {pfmt(alloc['weight'])} when VIXMO",
                "<=",
                f"{alloc['upper_bound']}",
            )

    symbols_table = Table(
        title="Configured symbols and target weights",
        box=box.SIMPLE_HEAVY,
        show_lines=True,
    )
    symbols_table.add_column("Symbol")
    symbols_table.add_column("Weight", justify="right")
    symbols_table.add_column("Call delta", justify="right")
    symbols_table.add_column("Call strike limit", justify="right")
    symbols_table.add_column("Call threshold", justify="right")
    symbols_table.add_column("HWM", justify="right")
    symbols_table.add_column("Put delta", justify="right")
    symbols_table.add_column("Put strike limit", justify="right")
    symbols_table.add_column("Put threshold", justify="right")
    for symbol, sconfig in config["symbols"].items():
        symbols_table.add_row(
            symbol,
            pfmt(sconfig["weight"]),
            ffmt(get_target_delta(config, symbol, "C")),
            dfmt(get_strike_limit(config, symbol, "C")),
            (
                f"{ffmt(get_write_threshold_sigma(config, symbol, 'C'))}σ"
                if get_write_threshold_sigma(config, symbol, "C")
                else pfmt(get_write_threshold_perc(config, symbol, "C"))
            ),
            str(maintain_high_water_mark(config, symbol)),
            ffmt(get_target_delta(config, symbol, "P")),
            dfmt(get_strike_limit(config, symbol, "P")),
            (
                f"{ffmt(get_write_threshold_sigma(config, symbol, 'P'))}σ"
                if get_write_threshold_sigma(config, symbol, "P")
                else pfmt(get_write_threshold_perc(config, symbol, "P"))
            ),
        )
    assert (
        round(
            sum([config["symbols"][s]["weight"] for s in config["symbols"].keys()]), 5
        )
        == 1.00000
    )

    tree = Tree(":control_knobs:")
    tree.add(Group(f":file_cabinet: Loaded from {config_path}", config_table))
    tree.add(Group(":yin_yang: Symbology", symbols_table))
    console.print(Panel(tree, title="Config"))

    if config.get("ib_insync", {}).get("logfile"):
        util.logToFile(config["ib_insync"]["logfile"])  # type: ignore

    def onConnected() -> None:
        portfolio_manager.manage()

    ib = IB()
    ib.connectedEvent += onConnected

    completion_future: Future[bool] = Future()
    portfolio_manager = PortfolioManager(config, ib, completion_future)

    probeContractConfig = config["watchdog"]["probeContract"]
    watchdogConfig = config.get("watchdog", {})
    del watchdogConfig["probeContract"]
    probeContract = Contract(
        secType=probeContractConfig["secType"],
        symbol=probeContractConfig["symbol"],
        currency=probeContractConfig["currency"],
        exchange=probeContractConfig["exchange"],
    )

    if not without_ibc:
        # TWS version is pinned to current stable
        ibc_config = config.get("ibc", {})
        # Remove any config params that aren't valid keywords for IBC
        ibc_keywords = {
            k: ibc_config[k] for k in ibc_config if k not in ["RaiseRequestErrors"]
        }
        ibc = IBC(1019, **ibc_keywords)

        ib.RaiseRequestErrors = ibc_config.get("RaiseRequestErrors", False)

        watchdog = Watchdog(ibc, ib, probeContract=probeContract, **watchdogConfig)
        watchdog.start()

        ib.run(completion_future)  # type: ignore
        watchdog.stop()
        ibc.terminate()
    else:
        ib.connect(
            watchdogConfig["host"],
            watchdogConfig["port"],
            clientId=watchdogConfig["clientId"],
            timeout=watchdogConfig["probeTimeout"],
            account=config["account"]["number"],
        )
        ib.run(completion_future)  # type: ignore
        ib.disconnect()



import math
from datetime import datetime
from operator import itemgetter
from typing import Any, Callable, Dict, List, Optional, Tuple

import ib_insync.objects
import ib_insync.ticker
from ib_insync import AccountValue, Order, PortfolioItem, TagValue, Ticker, util
from ib_insync.contract import Option

from thetagang.options import option_dte


def account_summary_to_dict(
    account_summary: List[AccountValue],
) -> Dict[str, AccountValue]:
    d: Dict[str, AccountValue] = dict()
    for s in account_summary:
        d[s.tag] = s
    return d


def portfolio_positions_to_dict(
    portfolio_positions: List[PortfolioItem],
) -> Dict[str, List[PortfolioItem]]:
    d: Dict[str, List[PortfolioItem]] = dict()
    for p in portfolio_positions:
        symbol = p.contract.symbol
        if symbol not in d:
            d[symbol] = []
        d[symbol].append(p)
    return d


def position_pnl(position: ib_insync.objects.PortfolioItem) -> float:
    return position.unrealizedPNL / abs(position.averageCost * position.position)


def get_short_positions(
    positions: List[PortfolioItem], right: str
) -> List[PortfolioItem]:
    return [
        p
        for p in positions
        if isinstance(p.contract, Option)
        and p.contract.right.upper().startswith(right.upper())
        and p.position < 0
    ]


def get_long_positions(
    positions: List[PortfolioItem], right: str
) -> List[PortfolioItem]:
    return [
        p
        for p in positions
        if isinstance(p.contract, Option)
        and p.contract.right.upper().startswith(right.upper())
        and p.position > 0
    ]


def count_short_option_positions(positions: List[PortfolioItem], right: str) -> int:
    return math.floor(-sum([p.position for p in get_short_positions(positions, right)]))


def weighted_avg_short_strike(
    positions: List[PortfolioItem], right: str
) -> Optional[float]:
    shorts = [
        (abs(p.position), p.contract.strike)
        for p in get_short_positions(positions, right)
    ]
    num = sum([p[0] * p[1] for p in shorts])
    den = sum([p[0] for p in shorts])
    if den > 0:
        return num / den


def weighted_avg_long_strike(
    positions: List[PortfolioItem], right: str
) -> Optional[float]:
    shorts = [
        (abs(p.position), p.contract.strike)
        for p in get_long_positions(positions, right)
    ]
    num = sum([p[0] * p[1] for p in shorts])
    den = sum([p[0] for p in shorts])
    if den > 0:
        return num / den


def count_long_option_positions(positions: List[PortfolioItem], right: str) -> int:
    return math.floor(sum([p.position for p in get_long_positions(positions, right)]))


def calculate_net_short_positions(positions: List[PortfolioItem], right: str) -> int:
    shorts = [
        (
            option_dte(p.contract.lastTradeDateOrContractMonth),
            p.contract.strike,
            p.position,
        )
        for p in get_short_positions(positions, right)
    ]
    longs = [
        (
            option_dte(p.contract.lastTradeDateOrContractMonth),
            p.contract.strike,
            p.position,
        )
        for p in get_long_positions(positions, right)
    ]
    shorts = sorted(shorts, key=itemgetter(0, 1), reverse=right.upper().startswith("P"))
    longs = sorted(longs, key=itemgetter(0, 1), reverse=right.upper().startswith("P"))

    def calc_net(short_dte: int, short_strike: float, short_position: float) -> float:
        for i in range(len(longs)):
            if short_position > -1:
                break
            (long_dte, long_strike, long_position) = longs[i]
            if long_position < 1:
                # ignore empty long positions
                continue
            if long_dte >= short_dte:
                if (
                    math.isclose(short_strike, long_strike)
                    or (right.upper().startswith("P") and long_strike >= short_strike)
                    or (right.upper().startswith("C") and long_strike <= short_strike)
                ):
                    if short_position + long_position > 0:
                        long_position = short_position + long_position
                        short_position = 0
                    else:
                        short_position += long_position
                        long_position = 0
            longs[i] = (long_dte, long_strike, long_position)
        return min([0.0, short_position])

    nets = [calc_net(*short) for short in shorts]

    return math.floor(-sum(nets))


def net_option_positions(
    symbol: str,
    portfolio_positions: Dict[str, List[PortfolioItem]],
    right: str,
    ignore_dte: Optional[int] = None,
) -> int:
    if symbol in portfolio_positions:
        return math.floor(
            sum(
                [
                    p.position
                    for p in portfolio_positions[symbol]
                    if isinstance(p.contract, Option)
                    and p.contract.right.upper().startswith(right.upper())
                    and option_dte(p.contract.lastTradeDateOrContractMonth) >= 0
                    and (
                        not ignore_dte
                        or option_dte(p.contract.lastTradeDateOrContractMonth)
                        > ignore_dte
                    )
                ]
            )
        )

    return 0


def wait_n_seconds(
    pred: Callable[[], bool],
    body: Callable[[float], Any],
    seconds_to_wait: int,
    started_at: Optional[datetime] = None,
) -> None:
    if not started_at:
        started_at = datetime.now()
    while pred():
        diff = datetime.now() - started_at
        remaining = seconds_to_wait - diff.seconds
        if not remaining or remaining <= 0 or math.isclose(remaining, 0.0):
            raise RuntimeError(
                "Exhausted retries waiting on predicate. This shouldn't happen."
            )
        body(remaining)


def get_higher_price(ticker: Ticker) -> float:
    # Returns the highest of either the option model price, the midpoint, or the
    # market price. The midpoint is usually a bit higher than the IB model's
    # pricing, but we want to avoid leaving money on the table in cases where
    # the spread might be messed up. This may in some cases make it harder for
    # orders to fill in a given day, but I think that's a reasonable tradeoff to
    # avoid leaving money on the table.
    if ticker.modelGreeks and ticker.modelGreeks.optPrice:
        return max([midpoint_or_market_price(ticker), ticker.modelGreeks.optPrice])
    return midpoint_or_market_price(ticker)


def get_lower_price(ticker: Ticker) -> float:
    # Same as get_highest_price(), except get the lower price instead.
    if ticker.modelGreeks and ticker.modelGreeks.optPrice:
        return min([midpoint_or_market_price(ticker), ticker.modelGreeks.optPrice])
    return midpoint_or_market_price(ticker)


def midpoint_or_market_price(ticker: Ticker) -> float:
    # As per the ib_insync docs, marketPrice returns the last price first, but
    # we often prefer the midpoint over the last price. This function pulls the
    # midpoint first, then falls back to marketPrice() if midpoint is nan.
    if util.isNan(ticker.midpoint()):
        if (
            util.isNan(ticker.marketPrice())
            and ticker.modelGreeks
            and ticker.modelGreeks.optPrice
        ):
            # Fallback to the model price if the greeks are available
            return ticker.modelGreeks.optPrice
        else:
            return ticker.marketPrice() if not util.isNan(ticker.marketPrice()) else 0.0

    return ticker.midpoint()


def get_target_dte(config: Dict[str, Any], symbol: str) -> int:
    if symbol in config["symbols"] and "dte" in config["symbols"][symbol]:
        return config["symbols"][symbol]["dte"]

    return config["target"]["dte"]


def get_target_delta(config: Dict[str, Any], symbol: str, right: str) -> float:
    p_or_c = "calls" if right.upper().startswith("C") else "puts"
    if (
        p_or_c in config["symbols"][symbol]
        and "delta" in config["symbols"][symbol][p_or_c]
    ):
        return config["symbols"][symbol][p_or_c]["delta"]
    if "delta" in config["symbols"][symbol]:
        return config["symbols"][symbol]["delta"]
    if p_or_c in config["target"]:
        return config["target"][p_or_c]["delta"]
    return config["target"]["delta"]


def get_cap_factor(config: Dict[str, Any], symbol: str) -> float:
    if (
        "calls" in config["symbols"][symbol]
        and "cap_factor" in config["symbols"][symbol]["calls"]
    ):
        return config["symbols"][symbol]["calls"]["cap_factor"]
    return config["write_when"]["calls"]["cap_factor"]


def get_cap_target_floor(config: Dict[str, Any], symbol: str) -> float:
    if (
        "calls" in config["symbols"][symbol]
        and "cap_target_floor" in config["symbols"][symbol]["calls"]
    ):
        return config["symbols"][symbol]["calls"]["cap_target_floor"]
    return config["write_when"]["calls"]["cap_target_floor"]


def get_strike_limit(
    config: Dict[str, Any], symbol: str, right: str
) -> Optional[float]:
    p_or_c = "calls" if right.upper().startswith("C") else "puts"
    if (
        p_or_c in config["symbols"][symbol]
        and "strike_limit" in config["symbols"][symbol][p_or_c]
    ):
        return config["symbols"][symbol][p_or_c]["strike_limit"]
    return None


def get_target_calls(
    config: Dict[str, Any], symbol: str, current_shares: int, target_shares: int
) -> int:
    cap_factor = get_cap_factor(config, symbol)
    cap_target_floor = get_cap_target_floor(config, symbol)
    min_uncovered = (target_shares * cap_target_floor) // 100
    max_covered = (current_shares * cap_factor) // 100
    total_coverable = current_shares // 100

    return max([0, math.floor(min([max_covered, total_coverable - min_uncovered]))])


def get_write_threshold_sigma(
    config: Dict[str, Any], symbol: Optional[str], right: str
) -> Optional[float]:
    p_or_c = "calls" if right.upper().startswith("C") else "puts"
    if symbol:
        if (
            p_or_c in config["symbols"][symbol]
            and "write_threshold_sigma" in config["symbols"][symbol][p_or_c]
        ):
            return config["symbols"][symbol][p_or_c]["write_threshold_sigma"]
        if "write_threshold_sigma" in config["symbols"][symbol]:
            return config["symbols"][symbol]["write_threshold_sigma"]
        # if there's a percentage-based threshold defined, we want to use that, so we return None here
        if (
            p_or_c in config["symbols"][symbol]
            and "write_threshold" in config["symbols"][symbol][p_or_c]
        ) or "write_threshold" in config["symbols"][symbol]:
            return None

    # check if there's a default value in constants
    if (
        p_or_c in config["constants"]
        and "write_threshold_sigma" in config["constants"][p_or_c]
    ):
        return config["constants"][p_or_c]["write_threshold_sigma"]
    if "write_threshold_sigma" in config["constants"]:
        return config["constants"]["write_threshold_sigma"]

    return None


def get_write_threshold_perc(
    config: Dict[str, Any], symbol: Optional[str], right: str
) -> float:
    p_or_c = "calls" if right.upper().startswith("C") else "puts"
    if symbol:
        if (
            p_or_c in config["symbols"][symbol]
            and "write_threshold" in config["symbols"][symbol][p_or_c]
        ):
            return config["symbols"][symbol][p_or_c]["write_threshold"]
        if "write_threshold" in config["symbols"][symbol]:
            return config["symbols"][symbol]["write_threshold"]

    # check if there's a default value in constants
    if (
        p_or_c in config["constants"]
        and "write_threshold" in config["constants"][p_or_c]
    ):
        return config["constants"][p_or_c]["write_threshold"]
    if "write_threshold" in config["constants"]:
        return config["constants"]["write_threshold"]
    return 0.0


def algo_params_from(params: List[str]) -> List[TagValue]:
    return [TagValue(p[0], p[1]) for p in params]


def get_minimum_credit(config: Dict[str, Any]) -> float:
    return config["orders"].get("minimum_credit", 0.0)


def maintain_high_water_mark(config: Dict[str, Any], symbol: str) -> bool:
    if (
        "calls" in config["symbols"][symbol]
        and "maintain_high_water_mark" in config["symbols"][symbol]["calls"]
    ):
        return config["symbols"][symbol]["calls"]["maintain_high_water_mark"]
    return config["roll_when"]["calls"]["maintain_high_water_mark"]


def get_max_dte_for(symbol: str, config: Dict[str, Any]) -> Optional[int]:
    if symbol == "VIX" and "max_dte" in config["vix_call_hedge"]:
        return config["vix_call_hedge"]["max_dte"]
    if symbol in config["symbols"] and "max_dte" in config["symbols"][symbol]:
        return config["symbols"][symbol]["max_dte"]

    return config["target"]["max_dte"]


def would_increase_spread(order: Order, updated_price: float) -> bool:
    return (
        order.action == "BUY"
        and updated_price < order.lmtPrice
        or order.action == "SELL"
        and updated_price > order.lmtPrice
    )


def can_write_when(
    config: Dict[str, Any], symbol: str, right: str
) -> Tuple[bool, bool]:
    p_or_c = "calls" if right.upper().startswith("C") else "puts"
    can_write_when_green = (
        config["symbols"][symbol][p_or_c]["write_when"]["green"]
        if p_or_c in config["symbols"][symbol]
        and "write_when" in config["symbols"][symbol][p_or_c]
        and "green" in config["symbols"][symbol][p_or_c]["write_when"]
        else config["write_when"][p_or_c]["green"]
    )
    can_write_when_red = (
        config["symbols"][symbol][p_or_c]["write_when"]["red"]
        if p_or_c in config["symbols"][symbol]
        and "write_when" in config["symbols"][symbol][p_or_c]
        and "red" in config["symbols"][symbol][p_or_c]["write_when"]
        else config["write_when"][p_or_c]["red"]
    )
    return (can_write_when_green, can_write_when_red)
